<?php
//------------------------------------------------------------------------------
//
// Network Configuration Script
//
//
// (c) 2019, by CYTEL Technology AG
//           www.cytel.de
//
// Editor:   Franz Berg
//           info@cytel.de
//
// Description:
//           Network configuration parser and setter.
//
//
//------------------------------------------------------------------------------

//----------------------------------------------
// Ini file
//----------------------------------------------

if( file_exists( __DIR__ . "/../access.ini" ) )  {
   $iniFile = parse_ini_file( __DIR__ . "/../access.ini" );
}
else if( file_exists( __DIR__ . "/../agw.ini" ) ) {
  $iniFile = parse_ini_file( __DIR__ . "/../agw.ini" );
}

//----------------------------------------------
// GLOBAL DEFINES
//----------------------------------------------

//define( "NETWORK_BOOT_SETTINGS",      "/opt/access6/config/networksettings.json"    );
define( "NETWORK_BOOT_SETTINGS",      $iniFile['root'] . "/config/networksettings.json" );
define( "NETWORK_INTERFACE_SETTINGS", "/etc/network/interfaces"                     );
define( "NETWORK_NTP_SETTINGS",       "/etc/ntp.conf"                               );
define( "DHCP_SERVICE_SETTINGS",      "/etc/dhcp/dhcpd.conf"                        );
define( "DHCP_INTERFACE_SETTINGS",    "/etc/default/isc-dhcp-server"                );
define( "DHCP_SYSTEMD_SETTINGS_PATH", "/etc/systemd/network/"                       );
define( "NETWORK_NTP_BACKUP_PATH",    $iniFile['root'] . "config/"                  );
define( "NETWORK_NTP_BACKUP_FILE",    $iniFile['root'] . "config/defaulttimeservers.json" );


//----------------------------------------------
// CLASS
//----------------------------------------------

class NetworkConfig
{
  //----------------------------------------------
  // MEMBER VARIABLES
  //----------------------------------------------

  protected $networkSettings; // Object to hold values read from networksettings.json.
  protected $exportSettings;  // Object to hold values to export.


  //----------------------------------------------
  // PUBLIC API
  //----------------------------------------------

  /**
   * Applies Configuration set in networksettings.json -> meant for the systemd service.
   * @return false|true
   */
  public function ApplyNetworkConfiguration()
  {
    // Check if configuration file exists.
    if( !file_exists( NETWORK_BOOT_SETTINGS  ) )  {
      die( "No network settings to apply.\n" ); // nothing to do
    }

    // Load and get all settings from configuration file.
    $this->networkSettings = $this->ReadNetworkSettings();

    // Check if networksettings are defined, else delete file and return null.
    if( empty( $this->networkSettings ) ) {
      unlink( NETWORK_BOOT_SETTINGS );
      die( "Given Netowkrsettings are null.\n" );
    }

    // Apply all settings.
    $this->SetNetworkNTPSettings();
    $this->RemoveNetworkInterfaceSettings();
    $this->SetNetworkInterfaceSettings();
    $this->SetDHCPInterfaceSettings();
    $this->SetDHCPSystemdSettings();
    $this->RemoveDHCPServiceSettings();
    $this->SetDHCPServiceSettings();

    // In the end, delete the file.
    unlink( NETWORK_BOOT_SETTINGS );

    // We're done, return true.
    return true;
  }

  /**
   * Gets the Networkconfiguration of the current system and returns as json.
   * @return false|object
   */
  public function GetNetworkConfiguration()
  {
    // Check if configuration file exists.
    if( !file_exists( NETWORK_BOOT_SETTINGS  ) )  {
      die( "No network settings to read.\n" ); // nothing to do
    }

    // Read all settings.
    $this->GetNetworkNTPSettings();
    $this->GetNetworkInterfaceSettings();
    $this->GetDHCPServiceSettings();

    // Return all settings.
    return json_encode( $this->exportSettings );
  }


  //----------------------------------------------
  // NETWORK SETTINGS
  //----------------------------------------------

  protected function ReadNetworkSettings()
  {
    // Prepare
    $jsonObject = null;
    $file       = file_get_contents( NETWORK_BOOT_SETTINGS );

    // Decode json file
    if( $file )  {
      $jsonObject = json_decode( $file, true );
    }

    // Return the object on success
    return $jsonObject;
  }


  //----------------------------------------------
  // NETWORK INTERFACE SETTINGS
  //----------------------------------------------

  protected function GetNetworkInterfaceSettings()
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Regex pattern for all needed lines from network conf.
    $lineMatchPattern  = '/^[ \t]*(dns-nameservers|address|netmask|gateway|noauto|auto|allow-|iface)\s*((?:(?!(?:dns-nameservers|address|netmask|gateway|noauto|auto|allow-|iface)).)*)(?<!\s)[ \t]/';
    $nicFile           = file_get_contents( NETWORK_INTERFACE_SETTINGS ); // Read data from network interfaces conf.
    $nicFileArray      = explode( "\n", $nicFile );                      // Convert into array.
    $lineIndexSettings = preg_grep( $lineMatchPattern, $nicFileArray );          // Get config from interfaces file.


    // =======================================================================
    //  READ NETWORK INTERFACES CONFIG
    // =======================================================================

    // Used to remember to which interface the settings are bound.
    $lastNetworkInterface = null;
    $index                = 0;

    // Get line -> value array and iterate through.
    foreach ( $lineIndexSettings as $key => $value ) {

      // Split at withespaces and don't allow empty indexes. iface eth0 inet static
      $parts = preg_split('#\s+#', $value, null, PREG_SPLIT_NO_EMPTY);

      if( preg_match('(iface)', $parts[0] ) === 1 ) {

        if( !empty( $lastNetworkInterface ) && $lastNetworkInterface !== $parts[1]) {
          ++$index;
        }

        $lastNetworkInterface                          = $parts[1];
        $this->exportSettings[ "interfaces" ][ $index ][ "adapter" ] = $parts[1];
        $this->exportSettings[ "interfaces" ][ $index ][ "type" ]    = $parts[3];
      }

      if( preg_match('(address)', $parts[0] ) === 1 ) {
        $this->exportSettings[ "interfaces" ][ $index ][ "ipv4" ] = $parts[1];
      }

      if( preg_match('(netmask)', $parts[0] ) === 1 ) {
        $this->exportSettings[ "interfaces" ][ $index ][ "networkmask" ] = $parts[1];
      }

      if( preg_match('(gateway)', $parts[0] ) === 1 ) {
        $this->exportSettings[ "interfaces" ][ $index ][ "gateway" ] = $parts[1];
      }

      if( preg_match('(dns-nameservers)', $parts[0] ) === 1 ) {
        $exportNameServers = array_slice( $parts, 1);
        $this->exportSettings[ "interfaces" ][ $index ][ "dns" ] = $exportNameServers;
      }
    }

  }

  protected function RemoveNetworkInterfaceSettings( )
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Read networksettings if not already done before.
    if( empty( $this->networkSettings ) ) {
      $this->ReadNetworkSettings();

      // Still null?
      if( empty( $this->networkSettings ) ) {
        return; // nothing to do.
      }
    }

    // Regex pattern for all needed lines from network conf.
    $lineMatchPattern     = '/^[ \t]*(dns-nameservers|address|netmask|gateway|noauto|auto|allow-|iface)\s*((?:(?!(?:dns-nameservers|address|netmask|gateway|noauto|auto|allow-|iface)).)*)(?<!\s)[ \t]/';

    // read file "/etc/network/interfaces"
    $nicFile              = file_get_contents(NETWORK_INTERFACE_SETTINGS );   // Read data from network interfaces conf.
    $nicFileArray         = explode( "\n", $nicFile );                        // Convert into array.
    $lineIndexSettings    = preg_grep( $lineMatchPattern, $nicFileArray );    // Get config from interfaces file.


    // =======================================================================
    //  REMOVE NIC
    // =======================================================================

    if( !key_exists( 'delete-nic', $this->networkSettings ) )  {
      return;
    }

    $adapterRemoveList = array_keys( $this->networkSettings['delete-nic'] );
    if( count( $adapterRemoveList ) > 0 )  {

      // get all valid source lines
      $lineIndexSettings = preg_grep( $lineMatchPattern, $nicFileArray );     // Get config from interfaces file.

      // search & remove adapter(s) given in section ['delete-nic']
      $removeLineFileArray = Array();
      $isGroupStarted = false;
      foreach( $lineIndexSettings as $key => $value )  {

        // in previous run an adapter had found, delete all from group
        if( $isGroupStarted == true )  {
          if( $value[ 0 ] != ' ' )  {   // group ends if there's no space at the beginning
            $isGroupStarted = false;
          }
          else  {                       // group continues
            array_push( $removeLineFileArray, $key );
          }
        }

        // search for adapter(s), if found => into deleteList
        for( $i = 0; $i < count( $adapterRemoveList ); $i++ )  {
          if( strpos( $value, $adapterRemoveList[ $i ] ) !== false
          && !empty($this->networkSettings['expertNetModus'])
          && $this->networkSettings['expertNetModus'][ $adapterRemoveList[ $i ] ] !== true )  {
            array_push( $removeLineFileArray, $key );
            $isGroupStarted = true;
          }
        }
      }//foreach

      // remove all lines that are marked to be deleted
      for( $i = count( $removeLineFileArray ) -1; $i >= 0; $i-- )  {
        array_splice( $nicFileArray, $removeLineFileArray[ $i ], 1 );
      }

      // write file
      $nicFile = implode( "\n", $nicFileArray ) . "\n";
      if( is_writable( NETWORK_INTERFACE_SETTINGS ) ) {
        file_put_contents( NETWORK_INTERFACE_SETTINGS, $nicFile );
      }
      else {
        echo "ERROR: File " . NETWORK_INTERFACE_SETTINGS . " is not writeable.\n";
      }


    }// remove

    // =======================================================================
    //  REMOVE NAMESERVER (dhclient.conf)
    // =======================================================================

    //find access generated settings and delete it ( nameservers settings are 3lines + comment id )
    shell_exec( "sudo sed -i '/# siedle access generated/,+3 d' /etc/dhcp/dhclient.conf" );
  }

  protected function SetNetworkInterfaceSettings( )
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Read networksettings if not already done before.
    if( empty( $this->networkSettings ) ) {
      $this->ReadNetworkSettings();

      // Still null?
      if( empty( $this->networkSettings ) ) {
        return; // nothing to do.
      }
    }

    // Regex pattern for all needed lines from network conf.
    $lineMatchPattern     = '/^[ \t]*(dns-nameservers|address|netmask|gateway|noauto|auto|allow-|iface)\s*((?:(?!(?:dns-nameservers|address|netmask|gateway|noauto|auto|allow-|iface)).)*)(?<!\s)[ \t]/';

    // Prepare.
    $nicFile              = file_get_contents(NETWORK_INTERFACE_SETTINGS );          // Read data from network interfaces conf.
    $lineIndexSettings    = preg_grep( $lineMatchPattern, explode( "\n", $nicFile )); // Get config from interfaces file.
    $interfaceConfigCache = array();                                                         // Cache data from data in reverse to prevent vector products.
    $containerCache       = array();                                                         // List to remember last container block.
    $lastContainer        = 0;                                                               // Key holder for the list.

    //find access generated settings and delete it ( nameservers settings are 3lines + comment id )
    shell_exec( "sudo sed -i '/# siedle access generated/,+1 d' /etc/dhcp/dhclient.conf" );
    // =======================================================================
    //  READ NETWORK INTERFACES CONFIG
    // =======================================================================

    // Get line -> value array and iterate through.
    foreach( $lineIndexSettings as $key => $value ) {

      // Split at withespaces and don't allow empty indexes. iface eth0 inet static
      $parts = preg_split( '#\s+#', $value, null, PREG_SPLIT_NO_EMPTY );

      // Match network initializers.
      if( preg_match( '(allow-hotplug|auto|noauto)', $parts[0] ) === 1 ) {

        // If nic has no name continue.
        if( !empty( $parts[1] ) ) {
            $interfaceConfigCache[ $parts[1] ][ "nicInitLine" ] = $key;
        }
      }

      // Match network dhcp option.
      if( preg_match('(iface)', $parts[0] ) === 1 ) {

        // We need to remember where the container begins!
        $lastContainer                    = $key;
        $containerCache[ $lastContainer ] = $parts[1]; // [eth0]

        // If nic has no name continue.
        if( !empty( $parts[3] ) ) { // [eth0][dhcp] = [static|dhcp]
          $interfaceConfigCache[ $parts[1] ][ "dhcp" ]             = $parts[3];
          $interfaceConfigCache[ $parts[1] ][ "nicContainerLine" ] = $key;
        }
      }

      // Match any container settings.
      if( preg_match('(address|netmask|gateway)', $parts[0] ) === 1 ) {

        // Now we can always get the nic by reading the container.
        $nic = $containerCache[ $lastContainer ];

        // If nic has no name continue.
        if( !empty( $parts[0] ) && !empty( $parts[1] ) ) {
          $interfaceConfigCache[ $nic ][ "container" ][ $parts[0] ] = $parts[1];
        }
      }

      // Nameserver specific settings
      if( preg_match('(dns-nameservers)', $parts[0] ) === 1 ) {

        // Now we can always get the nic by reading the container.
        $nic = $containerCache[ $lastContainer ];

        // If nic has no name continue.
        if( !empty( $parts[0] ) && !empty( $parts[1] ) && !empty( $parts[2] ) ) {
          $interfaceConfigCache[ $nic ][ "container" ][ $parts[0] ] = $parts[1] . " " . $parts[2];
        } elseif ( !empty( $parts[0] ) && !empty( $parts[1] ) ) {
          $interfaceConfigCache[ $nic ][ "container" ][ $parts[0] ] = $parts[1];
        }
      }

    }


    // =======================================================================
    //  WRITE NETWORK INTERFACES CONFIG
    // =======================================================================

    // Prepare.
    $fileContentArray = explode("\n", $nicFile ); // Handle file as Array.
    $fileContentItems = count($fileContentArray);         // Easy index handling.

    // First wipe the container lines of adapters we want to set up.
    foreach ( $this->networkSettings[ "interfaces" ] as $key => $value ) {

      // Prepare.
      $nicName          = $value[ "adapter" ];
      $nicInitLine      = empty( $interfaceConfigCache[ $nicName ][ "nicInitLine"      ] ) ? 0 : $interfaceConfigCache[ $nicName ][ "nicInitLine"      ];
      $nicContainerLine = empty( $interfaceConfigCache[ $nicName ][ "nicContainerLine" ] ) ? 0 : $interfaceConfigCache[ $nicName ][ "nicContainerLine" ];

      // Override if already initialized.
      if( !empty( $nicInitLine ) ) {
        $fileContentArray[ $nicInitLine ]      = "auto " . $nicName;
      }
      else {
        $fileContentArray[ $fileContentItems ] = "auto " . $nicName;
        ++$fileContentItems;
      }

      // Static/dhcp options.
      if( !empty( $nicContainerLine ) && $value[ "type" ] == "dhcp"  ) {
        $fileContentArray[ $nicContainerLine ] = "iface " . $nicName . " inet dhcp";
      }
      elseif( $value[ "type" ] == "dhcp" ) {
        $fileContentArray[ $fileContentItems ] = "iface " . $nicName . " inet dhcp";
        ++$fileContentItems;
        $fileContentArray[ $fileContentItems ] = ""; // Add whitespace.
        ++$fileContentItems;
      }
      else {
        $fileContentArray[ $fileContentItems ] = "iface " . $nicName . " inet static";
        ++$fileContentItems;

        if( !empty( $value[ "ipv4" ] ) ) {
          $fileContentArray[ $fileContentItems ] = "  address " . $value[ "ipv4" ];
          ++$fileContentItems;
        }

        if( !empty( $value[ "networkmask" ] ) ) {
          $fileContentArray[ $fileContentItems ] = "  netmask " . $value[ "networkmask" ];
          ++$fileContentItems;
        }

        if( !empty( $value[ "gateway" ] ) ) {
          $fileContentArray[ $fileContentItems ] = "  gateway " . $value[ "gateway" ];
          ++$fileContentItems;
        }

        if( !empty( $value[ "dns" ] ) ) {
          $dnsValue = implode( " ", $value[ "dns" ] );
          if( !empty( $dnsValue ) )  {
            $fileContentArray[ $fileContentItems ] = "  dns-nameservers " . $dnsValue;
            ++$fileContentItems;

            //save the nameserver also in /etc/dhcp/dhclient
            $currentNameServerConfig = PHP_EOL . "# siedle access generated" . PHP_EOL .
                                       "  prepend domain-name-servers " . $dnsValue . ";" . PHP_EOL;
            file_put_contents( "/etc/dhcp/dhclient.conf", $currentNameServerConfig,  FILE_APPEND  );
            shell_exec( "dhclient" );
          }
        }

        // Add whitespace.
        $fileContentArray[ $fileContentItems ] = "";
        ++$fileContentItems;

        // Cleanup trash lines.
        if( !empty( $nicContainerLine ) ) {

          // Remove dhcp/static setting.
          unset( $fileContentArray[ $nicContainerLine ] );
          ++$nicContainerLine;

          // Remove settings container.
          while( preg_match('/^\s+(\w)/', $fileContentArray[ $nicContainerLine ] ) === 1 ) {
            unset( $fileContentArray[ $nicContainerLine ] );
            ++$nicContainerLine;
          }
        }
      }
    } // foreach

    // Now clean and write the new file. Also don't forget the newline close inside the file.
    // BUG: this filters a duplicated '  netmask 255.255.255.0\n' line
    // $nicFile = implode( "\n", array_filter( array_unique ( $fileContentArray)  ) ) . "\n";
    $nicFile = implode( "\n", $fileContentArray ) . "\n";

    if( is_writable( NETWORK_INTERFACE_SETTINGS ) ) {
      file_put_contents( NETWORK_INTERFACE_SETTINGS, $nicFile );
    }
    else {
      echo "ERROR: File " . NETWORK_INTERFACE_SETTINGS . " is not writeable.\n";
    }


  }


  //----------------------------------------------
  // NETWORK NTP SETTINGS
  //----------------------------------------------

  protected function GetNetworkNTPSettings()
  {
    // =======================================================================
    //  PREQUISITES
    // =======================================================================
    if( !file_exists( NETWORK_NTP_SETTINGS ) )  {
      return;
    }

    // =======================================================================
    //  PREPARE
    // =======================================================================

    $ntpSettingsFile   = file_get_contents( NETWORK_NTP_SETTINGS );
    $ntpSettingsArray  = explode( "\n", $ntpSettingsFile );
    $ntpServerSettings = preg_grep( '/^[^#]*(pool)/', $ntpSettingsArray );
    $ntpServerBackup   = array();
    $ntpBackupExists   = file_exists( NETWORK_NTP_BACKUP_FILE );


    // =======================================================================
    //  GET NTP SERVER SETTINGS
    // =======================================================================

    // Check if backup exists.
    if( $ntpBackupExists ) {
      $ntpServerBackup = json_decode( file_get_contents( NETWORK_NTP_BACKUP_FILE ) );
    }

    // If backupservers are same like configured or backup doesn't exists its not custom conf.
    if ( $ntpServerSettings == $ntpServerBackup || !$ntpBackupExists ) {
      $this->exportSettings[ "ntpclient" ][ "custom" ] = false;
    } else {
      $this->exportSettings[ "ntpclient" ][ "custom" ] = true;
    }

    // Get nameservers to export.
    $ntpServerExport = array();
    foreach( $ntpServerSettings as $index => $ntpServer ) {
      $placeHolder = array();
      preg_match('/(?<=pool\s)\S*/', $ntpServer,  $placeHolder );
      if( !empty( $placeHolder[ 0 ] ) ) {
        array_push( $ntpServerExport, $placeHolder[0] );
      }
    }

    // Store in exportsettings object.
    $this->exportSettings["ntpclient"]["server"] = $ntpServerExport;
  }

  protected function SetNetworkNTPSettings()
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Read networksettings if not already done before.
    if( empty( $this->networkSettings ) ) {
      $this->ReadNetworkSettings();

      // Still null?
      if( empty( $this->networkSettings ) ) {
        return; // nothing to do.
      }
    }


    // =======================================================================
    //  WRITE NTP SETTINGS
    // =======================================================================

    $ntpSettingsFile   = file_get_contents( NETWORK_NTP_SETTINGS );
    $ntpSettingsArray  = explode( "\n", $ntpSettingsFile );
    $ntpServerSettings = preg_grep( '/^[^#]*(pool)/', $ntpSettingsArray );

    // Check if custom settings.
    if( $this->networkSettings[ "ntpclient" ][ "custom" ] === true ) {

      // Backup original config if no backup exists.
      if ( !file_exists( NETWORK_NTP_BACKUP_FILE ) ) {
        if( is_writable( NETWORK_NTP_BACKUP_PATH ) ) {
          file_put_contents( NETWORK_NTP_BACKUP_FILE, json_encode( array_values( $ntpServerSettings ) ) );
        }
        else {
          echo "ERROR: Filepath " . NETWORK_NTP_BACKUP_PATH . " is not writeable.\n";
        }
      }

      foreach( $this->networkSettings[ "ntpclient" ][ "server" ] as $ntpserver ) {
        array_push( $ntpSettingsArray, "pool ".  $ntpserver  . " iburst" );
      }
    }
    elseif( $this->networkSettings[ "ntpclient" ][ "custom" ] === false ) {

      // Check if we need to restore from backup.
      if( !file_exists( NETWORK_NTP_BACKUP_FILE ) ) {
        return; // Nothing to do, it is already default conf.
      }
      else {
        $nameServers = json_decode( file_get_contents( NETWORK_NTP_BACKUP_FILE ) );

        foreach( $nameServers as $ntpserver ) {
          array_push( $ntpSettingsArray, $ntpserver );
        }

        // Delete file after we restored our default timeservers.
        unlink( NETWORK_NTP_BACKUP_FILE );
      }
    }

    // =======================================================================
    //  CLEANUP NTP SETTINGS
    // =======================================================================

    // Prepare.
    $lastkey = null;

    // First compress whitespaces. -> Prevents growing whitespace configurations !!
    foreach( $ntpSettingsArray as $key => $value ) {

      if( $value === "" ) {

        if( ( $lastkey + 1 ) === $key ) {
          unset( $ntpSettingsArray[ $lastkey ] );
        }
        $lastkey = $key;
      }
    }

    // Then remove old ntp servers.
    foreach( $ntpServerSettings as $index => $ntpServer ) {
      unset( $ntpSettingsArray[ $index ] );
    }


    // =======================================================================
    //  WRITE NTP SETTINGS TO FILE
    // =======================================================================

    // Now glue the file together again and write it down.
    $ntpSettingsFile = implode( "\n", $ntpSettingsArray ) . "\n";
    if( is_writable( NETWORK_NTP_SETTINGS ) ) {
      file_put_contents( NETWORK_NTP_SETTINGS, $ntpSettingsFile );
    }
    else {
      echo "ERROR: File " . NETWORK_NTP_SETTINGS . " is not writeable.\n";
    }
  }


  //----------------------------------------------
  // DHCP SERVICE SETTINGS
  //----------------------------------------------

  protected function GetDHCPServiceSettings()
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    $dhcpSettingsFile  = file_get_contents( DHCP_SERVICE_SETTINGS );
    $dhcpSettingsArray = explode( "\n", $dhcpSettingsFile );
    $dhcpSettingsEnd   = count( $dhcpSettingsArray );

    // Options. Ignore '#' tagged ones.
    $dhcpSubnets = preg_grep( '/^[^#]*subnet\s(.*)netmask/', $dhcpSettingsArray );
    $blocknumber = 0;

    // Loop through our subnet config blocks.
    foreach( $dhcpSubnets as $index => $subnetDeclaration ) {

      // Extract ip address and netmask.
      $ipNetmaskExtraction = array();
      preg_match_all( '/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/', $subnetDeclaration, $ipNetmaskExtraction );

      if( count( $ipNetmaskExtraction[0] ) == 2 ) {
        $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "network"     ] = $ipNetmaskExtraction[0][0];
        $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "networkMask" ] = $ipNetmaskExtraction[0][1];
        $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "subnet-mask" ][ "number" ] = 1;
        $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "subnet-mask" ][ "type" ]   = "ip";
        $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "subnet-mask" ][ "value" ]  = $ipNetmaskExtraction[0][1];
      }

      // Prepare.
      $line        = $index;
      $blockended  = false;

      // Extract network adapter.
      while( $blockended == false ) {

        if( $line === $dhcpSettingsEnd ) {
          echo "WARNING: !!! MALFORMED SUBNET CONFIGURATION !!! \n";
          break;    // Malformed subnet!
        }

        $matchingOutput = array();

        if( preg_match('/(?<=\srange\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $range = preg_split('#\s+#', $matchingOutput[ 0 ], null, PREG_SPLIT_NO_EMPTY);
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "poolstart" ] = $range[0];
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "poolend" ]   = $range[1];
        }

        if( preg_match('/(?<=\sinterface\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "adapter" ] = $matchingOutput[0];
        }

        if( preg_match('/(?<=\stime-servers\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "time-servers" ][ "number" ]       = 4;
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "time-servers" ][ "type"   ]       = "iplist";
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "time-servers" ][ "value"  ]       = $matchingOutput[0];
        }

        if( preg_match('/(?<=\srouters\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "routers" ][ "number" ]           = 3;
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "routers" ][ "type"   ]           = "iplist";
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "routers" ][ "value"  ]           = $matchingOutput[0];
        }

        if( preg_match('/(?<=\sntp-servers\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "ntp-servers" ][ "number" ]       = 42;
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "ntp-servers" ][ "type"   ]       = "iplist";
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "ntp-servers" ][ "value"  ]       = $matchingOutput[0];
        }

        if( preg_match('/(?<=\stftp-server-name\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "tftp-server-name" ][ "number" ]  = 66;
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "tftp-server-name" ][ "type"   ]  = "iplist";
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "tftp-server-name" ][ "value"  ]  = $matchingOutput[0];
        }

        if( preg_match('/(?<=\sbootfile-name\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "bootfile-name" ][ "number" ]     = 67;
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "bootfile-name" ][ "type"   ]     = "text";
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "bootfile-name" ][ "value"  ]     = $matchingOutput[0];
        }

        if( preg_match('/(?<=\slog-servers\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $matchingOutput ) === 1 ) {
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "log-servers" ][ "number" ]       = 7;
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "log-servers" ][ "type"   ]       = "iplist";
          $this->exportSettings[ "dhcpserver" ][ $blocknumber ][ "options" ][ "log-servers" ][ "value"  ]       = $matchingOutput[0];
        }

        if( preg_match('/^\s*[}]+$/', $dhcpSettingsArray[ $line ] ) === 1 ) {
          $blockended = true;
          ++$blocknumber;
        }

        if( $dhcpSettingsEnd == $line ) {
          break; // Malformed configuration !!!
        }

        ++$line;
      }
    } // foreach dhcp subnet block

  }

  //----------------------------------------------
  // DHCP SETTINGS
  //----------------------------------------------

  protected function RemoveDHCPServiceSettings()
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Read networksettings if not already done before.
    if( empty( $this->networkSettings ) ) {
      $this->networkSettings = $this->ReadNetworkSettings();

      // Still null?
      if( empty( $this->networkSettings ) ) {
        return; // nothing to do.
      }
    }

    // Check if dhcp server settings are defined.
    if( empty( $this->networkSettings[ "delete-nic" ] ) ) {
      return; // nothing to do.
    }


    // =======================================================================
    //  READ DHCP SERVICE CONFIG
    // =======================================================================

    // Prepare.
    $dhcpSettingsFile  = file_get_contents( DHCP_SERVICE_SETTINGS );
    $dhcpSettingsArray = explode( "\n", $dhcpSettingsFile );
    $isInRemoveSection = false;

    for( $ln = 0; $ln < count( $dhcpSettingsArray ); $ln++ )  {

      // get current line
      $currentLine = trim( $dhcpSettingsArray[ $ln ] );

      // if previously the section has been marked as removed
      if( $isInRemoveSection == true )  {
        array_splice( $dhcpSettingsArray, $ln--, 1 );
        $pos = strpos( $currentLine, '}' );
        if( strpos( $currentLine, '}' ) !== false )  {
          $isInRemoveSection = false;
        }
      }

      // skip comments
      if( empty( $currentLine ) || $currentLine[ 0 ] == '#' )  {
        continue;
      }

      // analyse the subnet
      if( substr( $currentLine, 0, 6 ) == 'subnet' )  {
        $subNet = preg_split( '/[\s]+/', $currentLine );
        if( count( $subNet ) < 4 || $subNet[ 2 ] != 'netmask' )  // unknown format: must be "subnet 1.1.1.1 netmask 255.255.255.0 {"
          continue;

        foreach( $this->networkSettings[ "delete-nic" ] as $key => $value )  {
          if( $value['ipv4'] == $subNet[ 1 ] && $value['networkmask'] == $subNet[ 3 ] && $this->networkSettings['expertNetModus'][ $key ] !== true)  {
            $isInRemoveSection = true;
            array_splice( $dhcpSettingsArray, $ln--, 1 );
            break;
          }
        }
      }

    }//for $ln

    // Now clean and write the new file. Also don't forget the newline close inside the file.
    $dhcpSettingsFile = implode( "\n", $dhcpSettingsArray ) . "\n";

    if( is_writable( DHCP_SERVICE_SETTINGS ) ) {
      file_put_contents( DHCP_SERVICE_SETTINGS, $dhcpSettingsFile );
    }
    else {
      echo "ERROR: File " . DHCP_SERVICE_SETTINGS . " is not writeable.\n";
    }

  }

  protected function SetDHCPServiceSettings()
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Read networksettings if not already done before.
    if( empty( $this->networkSettings ) ) {
      $this->ReadNetworkSettings();

      // Still null?
      if( empty( $this->networkSettings ) ) {
        return; // nothing to do.
      }
    }

    // Check if dhcp server settings are defined.
    if( empty( $this->networkSettings[ "dhcpserver" ] ) ) {
      return; // nothing to do.
    }


    // =======================================================================
    //  READ DHCP SERVICE CONFIG
    // =======================================================================

    // Prepare.
    $dhcpSettingsFile  = file_get_contents( DHCP_SERVICE_SETTINGS );
    $dhcpSettingsArray = explode( "\n", $dhcpSettingsFile );
    $dhcpSettingsEnd   = count( $dhcpSettingsArray );

    // Options. Ignore '#' tagged ones.
    $dhcpAuthorativeOption = preg_grep( '/^[^#]*(authoritative;)/',    $dhcpSettingsArray );
    $dhcpSubnets           = preg_grep( '/^[^#]*subnet\s(.*)netmask/', $dhcpSettingsArray );
    $subnetAdapterLines    = array();

    // Convert $dhcpSubnets to [ip] <-> [line] storage.
    foreach( $dhcpSubnets as $key => $subnetIp ) {

      // Prepare.
      $line  = $key;
      $match = array();
      $found = false;

      // Extract network adapter.
      while( $found == false ) {

        if( $line === $dhcpSettingsEnd ) {
          echo "WARNING: !!! MALFORMED SUBNET CONFIGURATION !!! \n";
          break;    // Malformed subnet!
        }

        if( empty( $dhcpSettingsArray[ $line ] ) ) {
          ++$line;
          continue; // Already cleaned line.
        }

        $found = preg_match('/(?<=\sinterface\s).*?(?=;)/', $dhcpSettingsArray[ $line ], $match );
        ++$line;
      }

      // Bind networkadapter to subnet declaration line.
      if( $found ) {

        // Cleanup previous duplicate definition.
        if ( !empty( $subnetAdapterLines[ $match[ 0 ] ] ) && $subnetAdapterLines[ $match[ 0 ] ] < $key  ) {

          // Prepare.
          $line          = $subnetAdapterLines[ $match[ 0 ] ];
          $subnetCleaned = false;

          // Cleanup config.
          while( !$subnetCleaned ) {

            // If fileend got reached and no closing backet got found, its malformed !!!
            if( $line === $dhcpSettingsEnd ) {
              echo "WARNING: !!! CLEANED MALFORMED CONFIGURATION !!! \n";
              break;
            }

            if( empty( $dhcpSettingsArray[ $line ] ) ) {
              ++$line;
              continue; // Already cleaned line.
            }

            // Check for closing bracket, even honor malformed ones with spaces.
            if( preg_match('/^[^\S]*(\s*})/', $dhcpSettingsArray[ $line ], $output_array) ) {
              $subnetCleaned = true;
            }

            unset( $dhcpSettingsArray[ $line ] );

          }

          unset( $dhcpSubnets[ $subnetAdapterLines[ $match[ 0 ] ] ] );
        }

        $subnetAdapterLines[ $match[ 0 ] ] = $key;
      };
    }


    // =======================================================================
    //  CLEANUP OLD MATCHING SUBNET CONFIGURATIONS FOR ADAPTER
    // =======================================================================

    // Prepare.
    $lastkey = null;

    // First compress whitespaces. -> Prevents growing whitespace configurations !!
    foreach( $dhcpSettingsArray as $key => $value ) {

      if( $value === "" ) {

        if( ( $lastkey + 1 ) === $key ) {
          unset( $dhcpSettingsArray[ $lastkey ] );
        }
        $lastkey = $key;
      }
    }

    // Next match subnet adapters and remove their config blocks.
    foreach( $this->networkSettings[ "dhcpserver" ] as $key => $value ) {

      // If already a subnet for the given adapter is declared, then delete its config.
      if( is_array( $value ) && !empty( $subnetAdapterLines[ $value[ "adapter" ] ] ) )  {

        // Prepare.
        $line          = $subnetAdapterLines[ $value[ "adapter" ] ];
        $subnetCleaned = false;

        // Cleanup config.
        while( !$subnetCleaned ) {

          // If fileend got reached and no closing backet got found, its malformed !!!
          if( $line === $dhcpSettingsEnd ) {
            echo "WARNING: !!! CLEANED MALFORMED CONFIGURATION !!! \n";
            break;
          }

          if( empty( $dhcpSettingsArray[ $line ] ) ) {
            ++$line;
            continue; // Already cleaned line.
          }

          // Check for closing bracket, even honor malformed ones with spaces.
          if( preg_match('/^[^\S]*(\s*})/', $dhcpSettingsArray[ $line ], $output_array) ) {
            $subnetCleaned = true;
          }

          unset( $dhcpSettingsArray[ $line ] );
        }
      }
    }


    // =======================================================================
    //  WRITE NEW SUBNET CONFIGURATION
    // =======================================================================

    // Prepare.
    $hostname = trim( file_get_contents( "/etc/hostname" ), "\n" );

    // Add authorative option if missing.
    if ( empty( $dhcpAuthorativeOption ) ) {
      array_push( $dhcpSettingsArray, "authoritative;" );
      array_push( $dhcpSettingsArray, ""  );
    }

    // Set DHCP Server configurations.
    foreach( $this->networkSettings[ "dhcpserver" ] as $key => $value ) {

      // Check if its an array and values are set.
      if( is_array( $value )
        && !empty( $value[ "adapter"     ] )
        && !empty( $value[ "poolstart"   ] )
        && !empty( $value[ "poolend"     ] )
        && !empty( $value[ "networkmask" ] )
        && !empty( $value[ "network"     ] )
      ) {
        // Write block header.
        array_push( $dhcpSettingsArray, "subnet " . $value[ "network" ] . " netmask " . $value[ "networkmask" ] . "{" );

        // Block definitions.
        array_push( $dhcpSettingsArray, "  range "     . $value[ "poolstart" ] . " " . $value[ "poolend" ] . ";" );
        array_push( $dhcpSettingsArray, "  interface " . $value[ "adapter"   ] . ";"                             );

        // DHCP options.
        if( empty( $this->networkSettings[ "dhcpserver" ][ "options" ][ "domain-name" ] ) )  {
          array_push( $dhcpSettingsArray, "  option domain-name " . $hostname . ";" );
        }

        foreach( $this->networkSettings[ "dhcpserver" ][ $key ][ "options" ] as $option => $settings ) {
          if( empty( $settings[ "value" ] ) ) {
            continue; // ignore empty options
          }
          if( $option == "next-server" )  {
            array_push( $dhcpSettingsArray, " " . $option . " " . $settings[ "value" ] . ";" );
          }
          else  {
            array_push( $dhcpSettingsArray, "  option " . $option . " " . $settings[ "value" ] . ";" );
          }

        }

        // In the end the block closing bracket and a newline.
        array_push( $dhcpSettingsArray, "}" );
        array_push( $dhcpSettingsArray, ""  );
      }
    }


    // =======================================================================
    //  WRITE SUBNET CONFIG TO FILE
    // =======================================================================

    // Now clean and write the new file. Also don't forget the newline close inside the file.
    $dhcpSettingsFile = implode( "\n", $dhcpSettingsArray ) . "\n";

    if( is_writable( DHCP_SERVICE_SETTINGS ) ) {
      file_put_contents( DHCP_SERVICE_SETTINGS, $dhcpSettingsFile );
    }
    else {
      echo "ERROR: File " . DHCP_SERVICE_SETTINGS . " is not writeable.\n";
    }

  }


  //----------------------------------------------
  // DHCP INTERFACE SETTINGS"
  //----------------------------------------------

  protected function SetDHCPInterfaceSettings()
  {
    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Read networksettings if not already done before.
    if( empty( $this->networkSettings ) ) {
      $this->ReadNetworkSettings();

      // Still null?
      if( empty( $this->networkSettings ) ) {
        return; // nothing to do.
      }
    }

    // Check if dhcp server settings are defined.
    if( empty( $this->networkSettings[ "dhcpserver" ] ) ) {
      return; // nothing to do.
    }


    // =======================================================================
    //  GET DHCP INTERFACE SETTINGS
    // =======================================================================

    // Real config and get index where interfaces get set.
    $dhcpNicConfig      = file_get_contents( DHCP_INTERFACE_SETTINGS );
    $fileContentArray   = explode( "\n", $dhcpNicConfig );
    $interfacev4Indexes = preg_grep( '/(INTERFACESv4)/', $fileContentArray );
    $interfacev4Index   = 0;

    // Store interfaces to enable as dhcp server.
    $interfacesToEnable = array();

    // Store interfaces.
    foreach ( $this->networkSettings[ "dhcpserver" ] as $key => $value )  {
      if ( $value[ "enabled" ] === true ) {
        array_push( $interfacesToEnable, $value[ "adapter" ] );
      }
    }

    // Read indexes.
    $lastElement = end( $interfacev4Indexes );
    foreach ( $interfacev4Indexes as $key => $value )  {

      // Delete duplicate configuratins.
      if( $value != $lastElement ) {
        unset( $fileContentArray[ $key ] );
      } else {
        // Store position where the interface config key is stored.
        $interfacev4Index = $key;
      }
    }


    // =======================================================================
    //  WRITE DHCP INTERFACE SETTINGS
    // =======================================================================

    // Write data into the file array.
    $fileContentArray[ $interfacev4Index ] = 'INTERFACESv4="' . implode( ' ', $interfacesToEnable ) . '"';

    // Now glue the file together again and write it down.
    $dhcpNicConfig = implode( "\n", $fileContentArray ) . "\n";
    if( is_writable( DHCP_INTERFACE_SETTINGS ) ) {
      file_put_contents( DHCP_INTERFACE_SETTINGS, $dhcpNicConfig );
    }
    else {
      echo "ERROR: File " . DHCP_INTERFACE_SETTINGS . " is not writeable.\n";
    }
  }


  //----------------------------------------------
  // DHCP SYSTEMD SETTINGS
  //----------------------------------------------

  protected function SetDHCPSystemdSettings() {

    // =======================================================================
    //  PREPARE
    // =======================================================================

    // Read networksettings if not already done before.
    if( !empty( $this->networkSettings ) ) {
      $this->ReadNetworkSettings();

      // Still null?
      if( !empty( $this->networkSettings ) ) {
        return; // nothing to do.
      }
    }


    // =======================================================================
    //  WRITE DHCP SYSTEMD SETTINGS
    // =======================================================================

    // Write dhcp systemd settings if static and delete if set to dhcp.
    foreach ( $this->networkSettings[ "interfaces" ] as $key => $value )  {

      if( $value[ "type" ] === "static" && !empty( $value[ "adapter" ] ) ) {
        $filePath    = DHCP_SYSTEMD_SETTINGS_PATH . $value[ "adapter" ] . ".network";
        $fileContent = "[Match]\nName=" . $value[ "adapter" ] .  "\n\n[Network]\nDHCP=no\n\n";

        if( is_writable( $filePath ) ) {
          file_put_contents( $filePath, $fileContent );
        }
        else {
          echo "ERROR: File " . $filePath . " is not writeable.\n";
        }
      }
      elseif(  $value[ "type" ] === "dhcp" && !empty( $value[ "adapter" ] ) ) {
        $filePath    = DHCP_SYSTEMD_SETTINGS_PATH . $value[ "adapter" ] . ".network";

        if( is_writable( $filePath ) ) {
          unlink( $filePath );
        }
      }
    }

  }


  //----------------------------------------------
  // HELPER FUNCTIONS
  //----------------------------------------------

  protected function calculateSubnetAddress( $ip, $subnetmask )
  {
    $longIp     = ip2long( $ip );
    $longSubnet = ip2long( $subnetmask );
    return long2ip( $longIp & $longSubnet );
  }
}

// ##### USAGE EXAMPLE #####
$config = new NetworkConfig();
$config->ApplyNetworkConfiguration();
// $config->GetNetworkConfiguration();
